@@ -31,11 +31,7 @@ typedef struct _pthread_descr_struct *_pthread_descr;
 /* System specific semaphore definition.  */
 typedef struct
 {
-  struct
-  {
-    long int __status;
-    int __spinlock;
-  } __sem_lock;
+  struct _pthread_fastlock __sem_lock;
   int __sem_value;
   _pthread_descr __sem_waiting;
 } sem_t;
@@ -46,7 +42,7 @@ typedef struct
 #define SEM_FAILED	((sem_t *) 0)
 
 /* Maximum value the semaphore can have.  */
-#define SEM_VALUE_MAX 	((int) ((~0u) >> 1))
+#define SEM_VALUE_MAX 	(2147483647)
 
 
 __BEGIN_DECLS
@@ -68,13 +64,12 @@ extern int sem_close (sem_t *__sem) __THROW;
 extern int sem_unlink (__const char *__name) __THROW;
 
 /* Wait for SEM being posted.  */
-extern int sem_wait (sem_t *__sem) __THROW;
+extern int sem_wait (sem_t *__sem);
 
 #ifdef __USE_XOPEN2K
 /* Similar to `sem_wait' but wait only until ABSTIME.  */
 extern int sem_timedwait (sem_t *__restrict __sem,
-			  __const struct timespec *__restrict __abstime)
-     __THROW;
+			  __const struct timespec *__restrict __abstime);
 #endif
 
 /* Test whether SEM is posted.  */
