@@ -24,16 +24,14 @@
 #include "spinlock.h"
 #include "restart.h"
 
-#if !defined HAS_COMPARE_AND_SWAP || defined TEST_FOR_COMPARE_AND_SWAP
 static void __pthread_acquire(int * spinlock);
 
 static inline void __pthread_release(int * spinlock)
 {
   WRITE_MEMORY_BARRIER();
   *spinlock = __LT_SPINLOCK_INIT;
-  __asm __volatile ("" : "=m" (*spinlock) : "0" (*spinlock));
+  __asm __volatile ("" : "=m" (*spinlock) : "m" (*spinlock));
 }
-#endif
 
 
 /* The status field of a spinlock is a pointer whose least significant
@@ -87,13 +85,14 @@ void internal_function __pthread_lock(struct _pthread_fastlock * lock,
   spurious_wakeup_count = 0;
   spin_count = 0;
 
-again:
-
   /* On SMP, try spinning to get the lock. */
 
   if (__pthread_smp_kernel) {
     int max_count = lock->__spinlock * 2 + 10;
 
+    if (max_count > MAX_ADAPTIVE_SPIN_COUNT)
+      max_count = MAX_ADAPTIVE_SPIN_COUNT;
+
     for (spin_count = 0; spin_count < max_count; spin_count++) {
       if (((oldstatus = lock->__status) & 1) == 0) {
 	if(__compare_and_swap(&lock->__status, oldstatus, oldstatus | 1))
@@ -104,12 +103,17 @@ again:
 	  return;
 	}
       }
-      __asm __volatile ("" : "=m" (lock->__status) : "0" (lock->__status));
+#ifdef BUSY_WAIT_NOP
+      BUSY_WAIT_NOP;
+#endif
+      __asm __volatile ("" : "=m" (lock->__status) : "m" (lock->__status));
     }
 
     lock->__spinlock += (spin_count - lock->__spinlock) / 8;
   }
 
+again:
+
   /* No luck, try once more or suspend. */
 
   do {
@@ -126,7 +130,7 @@ again:
     }
 
     if (self != NULL) {
-      THREAD_SETMEM(self, p_nextlock, (pthread_descr) (oldstatus & ~1L));
+      THREAD_SETMEM(self, p_nextlock, (pthread_descr) (oldstatus));
       /* Make sure the store in p_nextlock completes before performing
          the compare-and-swap */
       MEMORY_BARRIER();
@@ -210,7 +214,7 @@ again:
       maxprio = thr->p_priority;
     }
     ptr = &(thr->p_nextlock);
-    thr = *ptr;
+    thr = (pthread_descr)((long)(thr->p_nextlock) & ~1L);
   }
 
   /* Remove max prio thread from waiting list. */
@@ -222,13 +226,13 @@ again:
        least significant bit is clear. */
     thr = (pthread_descr) (oldstatus & ~1L);
     if (! __compare_and_swap_with_release_semantics
-	    (&lock->__status, oldstatus, (long)(thr->p_nextlock)))
+	    (&lock->__status, oldstatus, (long)(thr->p_nextlock) & ~1L))
       goto again;
   } else {
     /* No risk of concurrent access, remove max prio thread normally.
        But in this case we must also flip the least significant bit
        of the status to mark the lock as released. */
-    thr = *maxptr;
+    thr = (pthread_descr)((long)*maxptr & ~1L);
     *maxptr = thr->p_nextlock;
 
     /* Ensure deletion from linked list completes before we
@@ -269,9 +273,7 @@ struct wait_node {
 };
 
 static long wait_node_free_list;
-#if !defined HAS_COMPARE_AND_SWAP || defined TEST_FOR_COMPARE_AND_SWAP
 static int wait_node_free_list_spinlock;
-#endif
 
 /* Allocate a new node from the head of the free list using an atomic
    operation, or else using malloc if that list is empty.  A fundamental
@@ -281,15 +283,6 @@ static int wait_node_free_list_spinlock;
 
 static struct wait_node *wait_node_alloc(void)
 {
-#if defined HAS_COMPARE_AND_SWAP
-  long oldvalue, newvalue;
-#endif
-
-#if defined TEST_FOR_COMPARE_AND_SWAP
-  if (!__pthread_has_cas)
-#endif
-#if !defined HAS_COMPARE_AND_SWAP || defined TEST_FOR_COMPARE_AND_SWAP
-  {
     struct wait_node *new_node = 0;
 
     __pthread_acquire(&wait_node_free_list_spinlock);
@@ -298,29 +291,12 @@ static struct wait_node *wait_node_alloc(void)
       wait_node_free_list = (long) new_node->next;
     }
     WRITE_MEMORY_BARRIER();
-    wait_node_free_list_spinlock = 0;
+    __pthread_release(&wait_node_free_list_spinlock);
 
     if (new_node == 0)
       return malloc(sizeof *wait_node_alloc());
 
     return new_node;
-  }
-#endif
-
-#if defined HAS_COMPARE_AND_SWAP
-  do {
-    oldvalue = wait_node_free_list;
-
-    if (oldvalue == 0)
-      return malloc(sizeof *wait_node_alloc());
-
-    /* Ensure we don't read stale next link through oldvalue pointer. */
-    READ_MEMORY_BARRIER();
-    newvalue = (long) ((struct wait_node *) oldvalue)->next;
-  } while (! __compare_and_swap(&wait_node_free_list, oldvalue, newvalue));
-
-  return (struct wait_node *) oldvalue;
-#endif
 }
 
 /* Return a node to the head of the free list using an atomic
@@ -328,33 +304,12 @@ static struct wait_node *wait_node_alloc(void)
 
 static void wait_node_free(struct wait_node *wn)
 {
-#if defined HAS_COMPARE_AND_SWAP
-  long oldvalue, newvalue;
-#endif
-
-#if defined TEST_FOR_COMPARE_AND_SWAP
-  if (!__pthread_has_cas)
-#endif
-#if !defined HAS_COMPARE_AND_SWAP || defined TEST_FOR_COMPARE_AND_SWAP
-  {
     __pthread_acquire(&wait_node_free_list_spinlock);
     wn->next = (struct wait_node *) wait_node_free_list;
     wait_node_free_list = (long) wn;
     WRITE_MEMORY_BARRIER();
-    wait_node_free_list_spinlock = 0;
+    __pthread_release(&wait_node_free_list_spinlock);
     return;
-  }
-#endif
-
-#if defined HAS_COMPARE_AND_SWAP
-  do {
-    oldvalue = wait_node_free_list;
-    wn->next = (struct wait_node *) oldvalue;
-    newvalue = (long) wn;
-    /* Ensure node contents are written before we swap it into the list. */
-    WRITE_MEMORY_BARRIER();
-  } while (! __compare_and_swap(&wait_node_free_list, oldvalue, newvalue));
-#endif
 }
 
 #if defined HAS_COMPARE_AND_SWAP
@@ -724,8 +679,7 @@ int __pthread_compare_and_swap(long * ptr, long oldval, long newval,
   return res;
 }
 
-/* This function is called if the inlined test-and-set
-   in __pthread_compare_and_swap() failed */
+#endif
 
 /* The retry strategy is as follows:
    - We test and set the spinlock MAX_SPIN_COUNT times, calling
@@ -764,5 +718,3 @@ static void __pthread_acquire(int * spinlock)
     }
   }
 }
-
-#endif
